from .manticore import Manticore
from ..core.plugin import InstructionCounter, Visited, Tracer, RecordSymbolicBranches


def native_main(args, _logger):
    env = {key: val for key, val in [env[0].split("=") for env in args.env]}

    m = Manticore(
        args.argv[0],
        argv=args.argv[1:],
        env=env,
        entry_symbol=args.entrysymbol,
        workspace_url=args.workspace,
        policy=args.policy,
        concrete_start=args.data,
        pure_symbolic=args.pure_symbolic,
    )

    # Default plugins for now.. FIXME REMOVE!
    m.register_plugin(InstructionCounter())
    m.register_plugin(Visited(args.coverage))
    m.register_plugin(Tracer())
    m.register_plugin(RecordSymbolicBranches())

    if args.names is not None:
        m.apply_model_hooks(args.names)

    if args.assertions:
        m.load_assertions(args.assertions)

    @m.init
    def init(state):
        for file in args.files:
            state.platform.add_symbolic_file(file)

    with m.kill_timeout():
        m.run()

    m.finalize()



from typing import Iterable

from ..platforms.platform import SyscallNotImplemented, unimplemented
from .linux_syscalls import amd64

import logging

logger = logging.getLogger(__name__)


class SyscallStubs:
    """
    DEVELOPING NEW SYSTEM CALLS:
    * Whenever you need to add proper support for a system call to Manticore, cut and paste the stub from this file into
      manticore/platforms/linux.py.
    * The stubs include the correct names and number of parameters, but do not include type information
    * Filling out docstrings is left as an exercise for the reader.
    * If self.default_to_fail is set, the stubs will do whatever the syscall would do when an error occurs
        * Otherwise, they'll try to pretend that they succeeded.
    * If the stub doesn't know what behavior to produce, it'll raise a SyscallNotImplemented exception
        * Examples include: When a call doesn't just return -1 on failure or 0 on success
    """

    def __init__(self, *, default_to_fail=False, parent=None):
        self.default_to_fail = default_to_fail
        self.parent = parent

    def __getattr__(self, item):
        logger.warning(
            f"Getting {item!r} attribute from {self.__class__}: "
            "System calls should be copied and pasted into linux.py, "
            "not implemented within the stub file. "
            "If you're seeing this message, you may have forgotten to do that."
        )
        return getattr(self.parent, item)

    def simple_returns(self):
        return -1 if self.default_to_fail else 0

    def complicated_success(self, num):
        if self.default_to_fail:
            return -1
        else:
            raise SyscallNotImplemented(num, amd64[num])

    def complicated_error(self, num):
        if self.default_to_fail:
            raise SyscallNotImplemented(num, amd64[num])
        else:
            return 0

    def complicated_both(self, num):
        if self.default_to_fail:
            raise SyscallNotImplemented(num, amd64[num])
        else:
            raise SyscallNotImplemented(num, amd64[num])

    @unimplemented
    def sys_accept4(self, fd, upeer_sockaddr, upeer_addrlen, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(288)

    @unimplemented
    def sys_add_key(self, _type, _description, _payload, plen) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(248)

    @unimplemented
    def sys_adjtimex(self, txc_p) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(159)

    @unimplemented
    def sys_alarm(self, seconds) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(37)

    @unimplemented
    def sys_bpf(self, cmd, attr, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(321)

    @unimplemented
    def sys_capget(self, header, dataptr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_capset(self, header, data) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_clock_getres(self, which_clock, tp) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_clock_nanosleep(self, which_clock, flags, rqtp, rmtp) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_error(230)

    @unimplemented
    def sys_clock_settime(self, which_clock, tp) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_connect(self, fd, uservaddr, addrlen) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_creat(self, pathname, mode) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(85)

    @unimplemented
    def sys_delete_module(self, name_user, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_dup3(self, oldfd, newfd, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(292)

    @unimplemented
    def sys_eventfd(self, count) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(284)

    @unimplemented
    def sys_faccessat(self, dfd, filename, mode) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fallocate(self, fd, mode, offset, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fanotify_init(self, flags, event_f_flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(300)

    @unimplemented
    def sys_fanotify_mark(self, fanotify_fd, flags, mask, dfd, pathname) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fchdir(self, fd) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fchmod(self, fd, mode) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fchmodat(self, dfd, filename, mode) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fchown(self, fd, user, group) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fchownat(self, dfd, filename, user, group, flag) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fcntl(self, fd, cmd, arg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(72)

    @unimplemented
    def sys_fdatasync(self, fd) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fgetxattr(self, fd, name, value, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(193)

    @unimplemented
    def sys_finit_module(self, fd, uargs, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_flistxattr(self, fd, list, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(196)

    @unimplemented
    def sys_flock(self, fd, cmd) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fremovexattr(self, fd, name) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fsetxattr(self, fd, name, value, size, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_fstatfs(self, fd, buf) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_futimesat(self, dfd, filename, utimes) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_get_mempolicy(self, policy, nmask, maxnode, addr, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_get_robust_list(self, pid, head_ptr, len_ptr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_error(274)

    @unimplemented
    def sys_getcpu(self, cpup, nodep, unused) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getdents64(self, fd, dirent, count) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(217)

    @unimplemented
    def sys_getgroups(self, gidsetsize, grouplist) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(115)

    @unimplemented
    def sys_getitimer(self, which, value) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getpeername(self, fd, usockaddr, usockaddr_len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getpgid(self, pid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(121)

    @unimplemented
    def sys_getpgrp(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(111)

    @unimplemented
    def sys_getppid(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(110)

    @unimplemented
    def sys_getpriority(self, which, who) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(140)

    @unimplemented
    def sys_getresgid(self, rgid, egid, sgid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getresuid(self, ruid, euid, suid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getrusage(self, who, ru) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getsid(self, pid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(124)

    @unimplemented
    def sys_getsockname(self, fd, usockaddr, usockaddr_len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getsockopt(self, fd, level, optname, optval, optlen) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_gettimeofday(self, tv, tz) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_getxattr(self, pathname, name, value, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(191)

    @unimplemented
    def sys_init_module(self, umod, len, uargs) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_inotify_add_watch(self, fd, pathname, mask) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(254)

    @unimplemented
    def sys_inotify_init(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(253)

    @unimplemented
    def sys_inotify_init1(self, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(294)

    @unimplemented
    def sys_inotify_rm_watch(self, fd, wd) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_io_cancel(self, ctx_id, iocb, result) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_io_destroy(self, ctx) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_io_getevents(self, ctx_id, min_nr, nr, events) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(208)

    @unimplemented
    def sys_io_setup(self, nr_events, ctxp) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_io_submit(self, ctx_id, nr, iocbpp) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(209)

    @unimplemented
    def sys_ioperm(self, _from, num, turn_on) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_ioprio_get(self, which, who) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(252)

    @unimplemented
    def sys_ioprio_set(self, which, who, ioprio) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_kcmp(self, pid1, pid2, type, idx1, idx2) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(312)

    @unimplemented
    def sys_kexec_file_load(self, kernel_fd, initrd_fd, cmdline_len, cmdline_ptr, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(320)

    @unimplemented
    def sys_kexec_load(self, entry, nr_segments, segments, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(246)

    @unimplemented
    def sys_keyctl(self, option, arg2, arg3, arg4, arg5) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(250)

    @unimplemented
    def sys_lchown(self, filename, user, group) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_lgetxattr(self, pathname, name, value, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(192)

    @unimplemented
    def sys_linkat(self, oldfd, oldname, newfd, newname, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_listxattr(self, pathname, list, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(194)

    @unimplemented
    def sys_llistxattr(self, pathname, list, size) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(195)

    @unimplemented
    def sys_lookup_dcookie(self, cookie64, buf, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(212)

    @unimplemented
    def sys_lremovexattr(self, pathname, name) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_lsetxattr(self, pathname, name, value, size, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mbind(self, start, len, mode, nmask, maxnode, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_membarrier(self, cmd, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(324)

    @unimplemented
    def sys_memfd_create(self, uname_ptr, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(319)

    @unimplemented
    def sys_migrate_pages(self, pid, maxnode, old_nodes, new_nodes) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(256)

    @unimplemented
    def sys_mincore(self, start, len, vec) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mknod(self, filename, mode, dev) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mknodat(self, dfd, filename, mode, dev) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mlock(self, start, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mlock2(self, start, len, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mlockall(self, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_modify_ldt(self, func, ptr, bytecount) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(154)

    @unimplemented
    def sys_mount(self, dev_name, dir_name, type, flags, data) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_move_pages(self, pid, nr_pages, pages, nodes, status, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mq_notify(self, mqdes, u_notification) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mq_open(self, u_name, oflag, mode, u_attr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(240)

    @unimplemented
    def sys_mq_timedreceive(self, mqdes, u_msg_ptr, msg_len, u_msg_prio, u_abs_timeout) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(243)

    @unimplemented
    def sys_mq_timedsend(self, mqdes, u_msg_ptr, msg_len, msg_prio, u_abs_timeout) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mq_unlink(self, u_name) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_mremap(self, addr, old_len, new_len, flags, new_addr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(25)

    @unimplemented
    def sys_msgctl(self, msqid, cmd, buf) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(71)

    @unimplemented
    def sys_msgget(self, key, msgflg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(68)

    @unimplemented
    def sys_msgrcv(self, msqid, msgp, msgsz, msgtyp, msgflg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(70)

    @unimplemented
    def sys_msgsnd(self, msqid, msgp, msgsz, msgflg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_msync(self, start, len, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_munlock(self, start, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_munlockall(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_name_to_handle_at(self, dfd, name, handle, mnt_id, flag) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_open_by_handle_at(self, dfd, name, handle, mnt_id, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(304)

    @unimplemented
    def sys_pause(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return -1

    @unimplemented
    def sys_perf_event_open(self, attr_uptr, pid, cpu, group_fd, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(298)

    @unimplemented
    def sys_personality(self, personality) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(135)

    @unimplemented
    def sys_pivot_root(self, new_root, put_old) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_poll(self, ufds, nfds, timeout_msecs) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(7)

    @unimplemented
    def sys_ppoll(self, ufds, nfds, tsp, sigmask, sigsetsize) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(271)

    @unimplemented
    def sys_prctl(self, option, arg2, arg3, arg4, arg5) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(157)

    @unimplemented
    def sys_preadv(self, fd, vec, vlen, pos_l, pos_h) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(295)

    @unimplemented
    def sys_process_vm_readv(self, pid, lvec, liovcnt, rvec, riovcnt, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(310)

    @unimplemented
    def sys_process_vm_writev(self, pid, lvec, liovcnt, rvec, riovcnt, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(311)

    @unimplemented
    def sys_ptrace(self, request, pid, addr, data) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(101)

    @unimplemented
    def sys_pwritev(self, fd, vec, vlen, pos_l, pos_h) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(296)

    @unimplemented
    def sys_quotactl(self, cmd, special, id, addr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_readahead(self, fd, offset, count) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_readlinkat(self, dfd, pathname, buf, bufsiz) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(267)

    @unimplemented
    def sys_reboot(self, magic1, magic2, cmd, arg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_recvfrom(self, fd, ubuf, size, flags, addr, addr_len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(45)

    @unimplemented
    def sys_recvmmsg(self, fd, mmsg, vlen, flags, timeout) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(299)

    @unimplemented
    def sys_recvmsg(self, fd, msg, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(47)

    @unimplemented
    def sys_remap_file_pages(self, start, size, prot, pgoff, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_removexattr(self, pathname, name) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_renameat(self, oldfd, oldname, newfd, newname) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_renameat2(self, olddfd, oldname, newdfd, newname, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_request_key(self, _type, _description, _callout_info, destringid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(249)

    @unimplemented
    def sys_restart_syscall(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_both(219)

    @unimplemented
    def sys_rseq(self, rseq, rseq_len, flags, sig) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_rt_sigpending(self, set, sigsetsize) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_rt_sigqueueinfo(self, pid, sig, uinfo) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_rt_sigsuspend(self, unewset, sigsetsize) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_rt_sigtimedwait(self, uthese, uinfo, uts, sigsetsize) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(128)

    @unimplemented
    def sys_rt_tgsigqueueinfo(self, tgid, pid, sig, uinfo) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_get_priority_max(self, policy) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(146)

    @unimplemented
    def sys_sched_get_priority_min(self, policy) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(147)

    @unimplemented
    def sys_sched_getaffinity(self, pid, len, user_mask_ptr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_getattr(self, pid, attr, size, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_getparam(self, pid, param) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_getscheduler(self, pid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(145)

    @unimplemented
    def sys_sched_rr_get_interval(self, pid, interval) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_setaffinity(self, pid, len, user_mask_ptr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_setattr(self, pid, attr, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_setparam(self, pid, param) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_setscheduler(self, pid, policy, param) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sched_yield(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_seccomp(self, op, flags, uargs) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_select(self, n, inp, outp, fd_setexp, tvp) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(23)

    @unimplemented
    def sys_semctl(self, semid, semnum, cmd, arg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(66)

    @unimplemented
    def sys_semget(self, key, nsems, semflg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(64)

    @unimplemented
    def sys_semop(self, semid, tsops, nsops) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_semtimedop(self, semid, tsops, nsops, timeout) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sendmmsg(self, fd, mmsg, vlen, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(307)

    @unimplemented
    def sys_sendmsg(self, fd, msg, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(46)

    @unimplemented
    def sys_sendto(self, fd, buff, len, flags, addr, addr_len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(44)

    @unimplemented
    def sys_set_mempolicy(self, mode, nmask, maxnode) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_set_robust_list(self, head, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_error(273)

    @unimplemented
    def sys_setdomainname(self, name, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setfsgid(self, gid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_both(123)

    @unimplemented
    def sys_setfsuid(self, uid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_both(122)

    @unimplemented
    def sys_setgid(self, gid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setgroups(self, gidsetsize, grouplist) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sethostname(self, name, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setitimer(self, which, value, ovalue) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setns(self, fd, nstype) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setpgid(self, pid, pgid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setpriority(self, which, who, niceval) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setregid(self, rgid, egid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setresgid(self, rgid, egid, sgid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setresuid(self, ruid, euid, suid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setreuid(self, ruid, euid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setrlimit(self, resource, rlim) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setsid(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(112)

    @unimplemented
    def sys_setsockopt(self, fd, level, optname, optval, optlen) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_settimeofday(self, tv, tz) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setuid(self, uid) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_setxattr(self, pathname, name, value, size, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_shmat(self, shmid, shmaddr, shmflg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(30)

    @unimplemented
    def sys_shmctl(self, shmid, cmd, buf) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(31)

    @unimplemented
    def sys_shmdt(self, shmaddr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_shmget(self, key, size, shmflg) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(29)

    @unimplemented
    def sys_shutdown(self, fd, how) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sigaltstack(self, uss, uoss) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_signalfd(self, ufd, user_mask, sizemask) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(282)

    @unimplemented
    def sys_socketpair(self, family, type, protocol, usockvec) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_splice(self, fd_in, off_in, fd_out, off_out, len, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(275)

    @unimplemented
    def sys_statfs(self, pathname, buf) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_swapoff(self, specialfile) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_swapon(self, specialfile, swap_flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_symlink(self, oldname, newname) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_symlinkat(self, oldname, newfd, newname) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sync(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sync_file_range(self, fd, offset, bytes, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_syncfs(self, fd) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_sysfs(self, option, arg1, arg2) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(139)

    @unimplemented
    def sys_sysinfo(self, info) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_syslog(self, type, buf, len) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(103)

    @unimplemented
    def sys_tee(self, fdin, fdout, len, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(276)

    @unimplemented
    def sys_tgkill(self, tgid, pid, sig) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_timer_create(self, which_clock, timer_event_spec, created_timer_id) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_timer_delete(self, timer_id) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_timer_getoverrun(self, timer_id) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(225)

    @unimplemented
    def sys_timer_gettime(self, timer_id, setting) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_timer_settime(self, timer_id, flags, new_setting, old_setting) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_timerfd_create(self, clockid, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(283)

    @unimplemented
    def sys_timerfd_gettime(self, ufd, otmr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_timerfd_settime(self, ufd, flags, utmr, otmr) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_times(self, info) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(100)

    @unimplemented
    def sys_tkill(self, pid, sig) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_truncate(self, path, length) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_unlinkat(self, dfd, pathname, flag) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_unshare(self, unshare_flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_ustat(self, dev, ubuf) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_utime(self, filename, times) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_utimensat(self, dfd, filename, utimes, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_utimes(self, filename, utimes) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_vhangup(self) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @unimplemented
    def sys_vmsplice(self, fd, iov, nr_segs, flags) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(278)

    @unimplemented
    def sys_wait4(self, upid, stat_addr, options, ru) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.complicated_success(61)

    @unimplemented
    def sys_waitid(self, which, upid, infop, options, ru) -> int:
        """AUTOGENERATED UNIMPLEMENTED STUB"""
        return self.simple_returns()

    @staticmethod
    def unimplemented_syscalls() -> Iterable[str]:
        import inspect

        return (
            x[0].split("sys_", 1)[1]
            for x in inspect.getmembers(SyscallStubs, predicate=inspect.isfunction)
            if x[0].startswith("sys_")
        )

    @staticmethod
    def print_unimplemented_syscalls() -> None:
        for syscall in SyscallStubs.unimplemented_syscalls():
            print(syscall)



import sys
import ctypes
import mmap as MMAP

mmap_function = None
munmap_function = None

# initialize global functions depending on platform


def get_libc():
    osname = sys.platform.lower()
    if osname.startswith("darwin"):
        filename = "libc.dylib"
    elif osname.startswith("linux"):
        filename = "libc.so.6"
    elif osname.startswith("netbsd"):
        filename = "libc.so"
    else:
        raise ValueError("Unsupported host OS: " + osname)

    return ctypes.cdll.LoadLibrary(filename)


libc = get_libc()

# void* mmap(void* addr, size_t len, int prot, int flags, int fd, off_t offset)
mmap_function = libc.mmap
mmap_function.restype = ctypes.c_void_p
mmap_function.argtype = [
    ctypes.c_void_p,
    ctypes.c_size_t,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_size_t,
]


# int munmap(void* addr, size_t len)
munmap_function = libc.munmap
munmap_function.restype = ctypes.c_int
munmap_function.argtype = [ctypes.c_void_p, ctypes.c_size_t]


def mmap(fd, offset, size):
    prot = MMAP.PROT_READ | MMAP.PROT_WRITE
    flags = MMAP.MAP_PRIVATE

    # When trying to map the contents of a file into memory, the offset must be a multiple of the page size (see
    # `man mmap`). So we need to align it before passing it to mmap(). Doing so also increases the size of the memory
    # area needed, so we need to account for that difference.
    aligned_offset = offset & ~0xFFF
    size += offset - aligned_offset

    if size & 0xFFF != 0:
        size = (size & ~0xFFF) + 0x1000
    assert size > 0

    result = mmap_function(0, size, prot, flags, fd, aligned_offset)
    assert result != ctypes.c_void_p(-1).value

    # Now when returning the pointer to the user, we need to skip the corrected offset so that the user doesn't end up
    # with a pointer to another region of the file than the one they requested.
    return ctypes.cast(result + offset - aligned_offset, ctypes.POINTER(ctypes.c_char))


def munmap(address, size):
    # When unmapping the memory area, we need to recover the pointer and the size that were given to mmap(). The
    # pointer can be recovered by aligning it on the page size. The size needs to be increased by that difference to
    # account for cases where the requested memory area was larger.
    address = ctypes.cast(address, ctypes.c_void_p).value
    aligned_address = address & ~0xFFF
    size += address - aligned_address
    assert size > 0
    aligned_address = ctypes.cast(aligned_address, ctypes.POINTER(ctypes.c_char))

    result = munmap_function(aligned_address, size)
    assert result == 0



import logging
import time
from typing import Any, Tuple, Dict

from capstone import CS_ARCH_ARM, CS_ARCH_X86, CS_MODE_32, CS_MODE_64

######################################################################
# Abstract classes for capstone/unicorn based cpus
# no emulator by default
from intervaltree import IntervalTree, Interval
from unicorn.unicorn_const import (
    UC_ARCH_X86,
    UC_MODE_64,
    UC_MODE_32,
    UC_PROT_NONE,
    UC_PROT_READ,
    UC_PROT_WRITE,
    UC_PROT_EXEC,
    UC_HOOK_MEM_READ_UNMAPPED,
    UC_HOOK_MEM_WRITE_UNMAPPED,
    UC_HOOK_MEM_FETCH_UNMAPPED,
    UC_HOOK_MEM_WRITE,
    UC_HOOK_INTR,
    UC_HOOK_INSN,
)
from unicorn import Uc, UcError
from unicorn.arm_const import *
from unicorn.x86_const import *

from ..core.smtlib import Operators, SelectedSolver, issymbolic
from ..native.memory import MemoryException

logger = logging.getLogger(__name__)


def convert_permissions(m_perms: str):
    """
    Converts a Manticore permission string into a Unicorn permission
    :param m_perms: Manticore perm string ('rwx')
    :return: Unicorn Permissions
    """
    permissions = UC_PROT_NONE
    if "r" in m_perms:
        permissions |= UC_PROT_READ
    if "w" in m_perms:
        permissions |= UC_PROT_WRITE
    if "x" in m_perms:
        permissions |= UC_PROT_EXEC
    return permissions


def hr_size(num, suffix="B") -> str:
    """
    Human-readable data size
    From https://stackoverflow.com/a/1094933
    :param num: number of bytes
    :param suffix: Optional size specifier
    :return: Formatted string
    """
    for unit in " KMGTPEZ":
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit if unit != " " else "", suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Y", suffix)


class ConcreteUnicornEmulator:
    """
    Helper class to emulate instructions in bulk via Unicorn.
    ---
    The regular Unicorn Emulator is used as a fallback for emulating single instructions that don't have their own
    implementations in Manticore. This Emulator is instead intended to completely replace Manticore's executor when
    operating on purely concrete data.

    To use the emulator, register a callback for the will_run event that calls `state.cpu.emulate_until` with an
    address at which it should switch back from Unicorn to Manticore. Passing 0 will result in the entire target being
    executed concretely.

    As a result of the concrete data requirement, this emulator is good for preloading concrete state, but typically
    should not be used once symbolic data is introduced. At time of writing, if you try emulate under Unicorn up until
    the point where symbolic data is introduced, switch to Manticore, fork states, then switch back, it *definitely*
    won't work.

    Only supports X86_64 for now.
    """

    def __init__(self, cpu):
        self._cpu = cpu
        self._mem_delta = {}
        self.flag_registers = {"CF", "PF", "AF", "ZF", "SF", "IF", "DF", "OF"}
        # Registers to ignore when translating manticore context to unicorn
        self.ignore_registers = {"MXCSR_MASK"}
        self.write_backs_disabled = False
        self._stop_at = None
        # Holds key of range (addr, addr + size) and value of permissions
        # Key doesn't include permissions because unmap doesn't care about permissions
        self.already_mapped: IntervalTree = IntervalTree()

        cpu.subscribe("did_write_memory", self.write_back_memory)
        cpu.subscribe("did_write_register", self.write_back_register)
        cpu.subscribe("did_set_descriptor", self.update_segment)
        cpu.subscribe("did_map_memory", self.map_memory_callback)
        cpu.subscribe("did_unmap_memory", self.unmap_memory_callback)
        cpu.subscribe("did_protect_memory", self.protect_memory_callback)

        if self._cpu.arch == CS_ARCH_X86:
            self._uc_arch = UC_ARCH_X86
            self._uc_mode = {CS_MODE_32: UC_MODE_32, CS_MODE_64: UC_MODE_64}[self._cpu.mode]
        else:
            raise NotImplementedError(f"Unsupported architecture: {self._cpu.arch}")

        self.reset()

        self._emu.hook_add(UC_HOOK_MEM_READ_UNMAPPED, self._hook_unmapped)
        self._emu.hook_add(UC_HOOK_MEM_WRITE_UNMAPPED, self._hook_unmapped)
        self._emu.hook_add(UC_HOOK_MEM_FETCH_UNMAPPED, self._hook_unmapped)
        self._emu.hook_add(UC_HOOK_MEM_WRITE, self._hook_write_mem)
        self._emu.hook_add(UC_HOOK_INTR, self._interrupt)
        self._emu.hook_add(UC_HOOK_INSN, self._hook_syscall, arg1=UC_X86_INS_SYSCALL)
        self._emu.hook_add(UC_HOOK_INSN, self._hook_cpuid, arg1=UC_X86_INS_CPUID)

        self.registers = set(self._cpu.canonical_registers)
        # The last 8 canonical registers of x86 are individual flags; replace with the eflags
        self.registers -= self.flag_registers
        self.registers.add("EFLAGS")

        self.load_state_from_manticore()

    def reset(self):
        self._emu = Uc(self._uc_arch, self._uc_mode)
        self._to_raise = None

    def copy_memory(self, address: int, size: int):
        """
        Copy the bytes from address to address+size into Unicorn
        Used primarily for copying memory maps
        :param address: start of buffer to copy
        :param size: How many bytes to copy
        """
        start_time = time.time()
        map_bytes = self._cpu._raw_read(address, size, force=True)
        self._emu.mem_write(address, map_bytes)
        if time.time() - start_time > 3:
            logger.info(
                f"Copying {hr_size(size)} map at {address:#x} took {time.time() - start_time} seconds"
            )

    def load_state_from_manticore(self) -> None:
        for reg in self.registers:
            # Ignore registers that aren't supported by unicorn
            if reg in self.ignore_registers:
                continue
            val = self._cpu.read_register(reg)
            if issymbolic(val):
                from ..native.cpu.abstractcpu import ConcretizeRegister

                raise ConcretizeRegister(
                    self._cpu, reg, "Concretizing for emulation.", policy="ONE"
                )

            if reg in {"FS", "GS"}:
                if reg == "FS" and val in self._cpu._segments:
                    base, limit, perms = self._cpu._segments[val]
                    self.update_segment(val, base, limit, perms)
                    continue
                logger.debug("Writing {val} into {reg}")
                self.msr_write(reg, val)
                continue

            logger.debug("Writing {val} into {reg}")
            self._emu.reg_write(self._to_unicorn_id(reg), val)

        for m in self._cpu.memory.maps:
            self.map_memory_callback(m.start, len(m), m.perms, m.name, 0, m.start)

    def map_memory_callback(
        self, address: int, size: int, perms: str, name: str, offset: int, result: int
    ) -> None:
        """
        Catches did_map_memory and copies the mapping into Manticore
        """
        begin = address
        end = address + size
        perms_value = convert_permissions(perms)
        # Check for exact match
        # Overlap match
        if (
            Interval(begin, end, perms_value) not in self.already_mapped
            and not self.already_mapped.overlaps(begin, end)
            and not self.already_mapped.envelop(begin, end)
        ):
            logger.info(
                " ".join(
                    (
                        "Mapping Memory @",
                        hex(address),
                        ":",
                        hex(address + size),
                        hr_size(size),
                        "-",
                        perms,
                        "-",
                        f"{name}:{offset:#x}" if name else "",
                        "->",
                        hex(result),
                    )
                )
            )
            self._emu.mem_map(begin, size, perms_value)
            self.already_mapped[begin:end] = perms_value
        logger.debug(
            " ".join(
                (
                    "Copying Memory @",
                    hex(address),
                    hr_size(size),
                    "-",
                    perms,
                    "-",
                    f"{name}:{offset:#x}" if name else "",
                    "->",
                    hex(result),
                )
            )
        )
        self.copy_memory(address, size)
        self.protect_memory_callback(address, size, perms)

    def unmap_memory_callback(self, start, size):
        """Unmap Unicorn maps when Manticore unmaps them"""

        # Need this check because our memory events are leaky to internal implementation details
        end = start + size
        parent_map = self.already_mapped.overlap(start, end)
        # Only unmap whole original maps
        if (
            len(parent_map) == 1
            and list(parent_map)[0].begin == start
            and list(parent_map)[0].end == end
        ):
            mask = (1 << 12) - 1
            if (start & mask) != 0:
                logger.error("Memory to be unmapped is not aligned to a page")

            if (size & mask) != 0:
                size = ((size >> 12) + 1) << 12
                logger.warning("Forcing unmap size to align to a page")

            logger.info(f"Unmapping memory from {start:#x} to {start+size:#x}")

            self._emu.mem_unmap(start, size)
            self.already_mapped.remove_overlap(start, start + size)
        else:
            logger.debug(
                f"Not unmapping because bounds ({start:#x} - {start+size:#x}) are enveloped in existing map:"
            )
            logger.debug(f"\tParent map(s) {parent_map}")

    def protect_memory_callback(self, start, size, perms):
        """Set memory protections in Unicorn correctly"""
        logger.debug(f"Changing permissions on {start:#x}:{start+size:#x} to '{perms}'")
        self._emu.mem_protect(start, size, convert_permissions(perms))

    def get_unicorn_pc(self):
        """Get the program counter from Unicorn regardless of architecture.
        Legacy method, since this module only works on x86."""
        if self._cpu.arch == CS_ARCH_ARM:
            return self._emu.reg_read(UC_ARM_REG_R15)
        elif self._cpu.arch == CS_ARCH_X86:
            if self._cpu.mode == CS_MODE_32:
                return self._emu.reg_read(UC_X86_REG_EIP)
            elif self._cpu.mode == CS_MODE_64:
                return self._emu.reg_read(UC_X86_REG_RIP)

    def _hook_syscall(self, uc, data):
        """
        Unicorn hook that transfers control to Manticore so it can execute the syscall
        """
        logger.debug(
            f"Stopping emulation at {uc.reg_read(self._to_unicorn_id('RIP')):#x} to perform syscall"
        )
        self.sync_unicorn_to_manticore()
        from ..native.cpu.abstractcpu import Syscall

        self._to_raise = Syscall()
        uc.emu_stop()

    def _hook_cpuid(self, uc, data):
        """
        Unicorn hook that uses Manticore's semantics for cpuid
        """
        logger.debug(f"Hooking CPUID instruction {uc.reg_read(self._to_unicorn_id('RIP')):#x}")
        if self._cpu.mode == CS_MODE_32:
            pc = uc.reg_read(UC_X86_REG_EIP)
        elif self._cpu.mode == CS_MODE_64:
            pc = uc.reg_read(UC_X86_REG_RIP)
        eax = uc.reg_read(UC_X86_REG_EAX)
        ecx = uc.reg_read(UC_X86_REG_ECX)

        from ..native.cpu.x86 import X86Cpu

        eax, ebx, ecx, edx = X86Cpu.CPUID_helper(pc, eax, ecx)

        uc.reg_write(UC_X86_REG_EAX, eax)
        uc.reg_write(UC_X86_REG_EBX, ebx)
        uc.reg_write(UC_X86_REG_ECX, ecx)
        uc.reg_write(UC_X86_REG_EDX, edx)

        return 1

    def _hook_write_mem(self, uc, _access, address: int, size: int, value: int, _data) -> bool:
        """
        Captures memory written by Unicorn
        """
        self._mem_delta[address] = (value, size)
        return True

    def _hook_unmapped(self, uc, access, address, size, value, _data) -> bool:
        """
        We hit an unmapped region; map it into unicorn.
        """
        try:
            self.sync_unicorn_to_manticore()
            logger.warning(f"Encountered an operation on unmapped memory at {address:#x}")
            m = self._cpu.memory.map_containing(address)
            self.copy_memory(m.start, m.end - m.start)
        except MemoryException as e:
            logger.error(f"Failed to map memory {address:#x}-{address+size:#x}, ({access}): {e}")
            self._to_raise = e
            self._should_try_again = False
            return False

        self._should_try_again = True
        return False

    def _interrupt(self, uc, number: int, _data) -> bool:
        """
        Handle software interrupt (SVC/INT)
        """
        logger.info(f"Caught interrupt: {number}")
        from ..native.cpu.abstractcpu import Interruption  # prevent circular imports

        self._to_raise = Interruption(number)
        return True

    def _to_unicorn_id(self, reg_name: str) -> int:
        if self._cpu.arch == CS_ARCH_ARM:
            return globals()["UC_ARM_REG_" + reg_name]
        elif self._cpu.arch == CS_ARCH_X86:
            # TODO(yan): This needs to handle AF register
            custom_mapping = {"PC": "RIP", "STACK": "RSP", "FRAME": "RBP", "FS_BASE": "FS_BASE"}
            try:
                return globals()["UC_X86_REG_" + custom_mapping.get(reg_name, reg_name)]
            except KeyError:
                logger.error("Can't find register UC_X86_REG_%s", str(reg_name))
                raise

        else:
            # TODO(yan): raise a more appropriate exception
            raise TypeError

    def emulate(self, instruction) -> None:
        """
        Wrapper that runs the _step function in a loop while handling exceptions
        """

        # The emulation might restart if Unicorn needs to bring in a memory map
        # or bring a value from Manticore state.
        while True:

            # Try emulation
            self._should_try_again = False
            self._to_raise = None

            self._step(instruction)

            if not self._should_try_again:
                break

    def _step(self, instruction, chunksize: int = 0) -> None:
        """
        Execute a chunk fo instructions starting from instruction
        :param instruction: Where to start
        :param chunksize: max number of instructions to execute. Defaults to infinite.
        """

        try:
            pc = self._cpu.PC
            m = self._cpu.memory.map_containing(pc)
            if self._stop_at:
                logger.info(f"Emulating from {pc:#x} to  {self._stop_at:#x}")
            self._emu.emu_start(pc, m.end if not self._stop_at else self._stop_at, count=chunksize)
        except UcError:
            # We request re-execution by signaling error; if we we didn't set
            # _should_try_again, it was likely an actual error
            if not self._should_try_again:
                raise

        if self._should_try_again:
            return

        # self.sync_unicorn_to_manticore()
        self._cpu.PC = self.get_unicorn_pc()
        if self._cpu.PC == self._stop_at:
            logger.info("Reached emulation target, switching to Manticore mode")
            self.sync_unicorn_to_manticore()
            self._stop_at = None
            self.write_backs_disabled = True

        # Raise the exception from a hook that Unicorn would have eaten
        if self._to_raise:
            from ..native.cpu.abstractcpu import Syscall

            if type(self._to_raise) is Syscall:
                # NOTE: raises Syscall within sem_SYSCALL
                # NOTE: Need to call syscall semantic function due to
                # @instruction around SYSCALL
                self._cpu.sem_SYSCALL()
            logger.info(f"Raising {self._to_raise}")
            raise self._to_raise

        logger.info(f"Exiting Unicorn Mode at {self._cpu.PC:#x}")
        return

    def sync_unicorn_to_manticore(self):
        """
        Copy registers and written memory back into Manticore
        """
        self.write_backs_disabled = True
        for reg in self.registers:
            # Ignore registers that aren't supported by unicorn
            if reg in self.ignore_registers:
                continue
            val = self._emu.reg_read(self._to_unicorn_id(reg))
            self._cpu.write_register(reg, val)
        if len(self._mem_delta) > 0:
            logger.debug(f"Syncing {len(self._mem_delta)} writes back into Manticore")
        for location in self._mem_delta:
            value, size = self._mem_delta[location]
            self._cpu.write_int(location, value, size * 8)
        self.write_backs_disabled = False
        self._mem_delta = {}

    def write_back_memory(self, where, expr, size):
        """Copy memory writes from Manticore back into Unicorn in real-time"""
        if self.write_backs_disabled:
            return
        if type(expr) is bytes:
            self._emu.mem_write(where, expr)
        else:
            if issymbolic(expr):
                data = [
                    Operators.CHR(Operators.EXTRACT(expr, offset, 8))
                    for offset in range(0, size, 8)
                ]
                concrete_data = []
                for c in data:
                    if issymbolic(c):
                        c = chr(
                            SelectedSolver.instance().get_value(self._cpu.memory.constraints, c)
                        )
                    concrete_data.append(c)
                data = concrete_data
            else:
                data = [
                    Operators.CHR(Operators.EXTRACT(expr, offset, 8))
                    for offset in range(0, size, 8)
                ]
            logger.debug(f"Writing back {hr_size(size // 8)} to {hex(where)}: {data}")
            # TODO - the extra encoding is to handle null bytes output as strings when we concretize. That's probably a bug.
            self._emu.mem_write(
                where, b"".join(b.encode("utf-8") if type(b) is str else b for b in data)
            )

    def write_back_register(self, reg, val):
        """Sync register state from Manticore -> Unicorn"""
        # Ignore registers that aren't supported by unicorn
        if reg in self.ignore_registers:
            return
        if self.write_backs_disabled:
            return
        if issymbolic(val):
            logger.warning("Skipping Symbolic write-back")
            return
        if reg in self.flag_registers:
            self._emu.reg_write(self._to_unicorn_id("EFLAGS"), self._cpu.read_register("EFLAGS"))
            return
        self._emu.reg_write(self._to_unicorn_id(reg), val)

    def update_segment(self, selector, base, size, perms):
        """Only useful for setting FS right now."""
        logger.debug("Updating selector %s to 0x%02x (%s bytes) (%s)", selector, base, size, perms)
        self.write_back_register("FS", selector)
        self.write_back_register("FS_BASE", base)
        self.msr_write("FS", base)

    def msr_write(self, reg, data):
        """
        set the hidden descriptor-register fields to the given address.
        This enables referencing the fs segment on x86-64.

        https://wiki.osdev.org/SWAPGS
        """
        magic = {"FS": 0xC0000100, "GS": 0xC0000101}
        return self._emu.msr_write(magic[reg], data)



from .expression import Expression, Bool, BitVec, Array, BitVecConstant, issymbolic  # noqa
from .constraints import ConstraintSet  # noqa
from .solver import *  # noqa
from . import operators as Operators  # noqa

import logging

logger = logging.getLogger(__name__)



from ..core.state import StateBase
from ..core.plugin import StateDescriptor


class State(StateBase):
    def execute(self):
        super().execute()
        return self._platform.execute()

    def _update_state_descriptor(self, descriptor: StateDescriptor, *args, **kwargs):
        """
        Called on execution_intermittent to update the descriptor for this state.
        This one should apply any EVM-specific information to the descriptor.

        :param descriptor: StateDescriptor for this state
        """
        super()._update_state_descriptor(descriptor, *args, **kwargs)
        descriptor.pc = (self.platform.current_vm.address, self.platform.current_vm.pc)



from .platform import Platform
from ..wasm.structure import (
    ModuleInstance,
    Store,
    FuncAddr,
    HostFunc,
    Stack,
    ProtoFuncInst,
    MemInst,
    MemAddr,
    GlobalInst,
    GlobalAddr,
    TableInst,
    TableAddr,
    ExternVal,
    Module,
)
from ..wasm.types import Trap, TypeIdx, TableType, MemoryType, GlobalType, MissingExportException
from ..core.state import TerminateState
from ..core.smtlib import ConstraintSet
from functools import partial
import typing
import logging
import os

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


def stub(arity, _state, *args):
    """Default function used for hostfunc calls when a proper import wasn't provided"""
    logger.info("Called stub function with args: %s", args)
    return [0 for _ in range(arity)]  # TODO: Return symbolic values


class WASMWorld(Platform):
    """Manages global environment for a WASM state. Analagous to EVMWorld."""

    def __init__(self, filename, name="self", **kwargs):
        """
        :param filename: The WASM module to execute
        :param kwargs: Accepts "constraints" to pass in an initial ConstraintSet
        """
        super().__init__(filename, **kwargs)
        #: Initial set of constraints
        self.constraints = kwargs.get("constraints", ConstraintSet())
        #: Prevents users from calling run without instantiating the module
        self.instantiated = []
        #: Backing store for functions, memories, tables, and globals
        self.store = Store()

        self.modules = []
        self.module_names = {}
        self.manual_exports = {}
        self.default_module = name
        self.register_module(name, filename)

        #: Stores numeric values, branch labels, and execution frames
        self.stack = Stack()

        #: Stores concretized information used to advise execution of the next instruction.
        self.advice = None

        self.forward_events_from(self.stack)
        self.forward_events_from(self.instance)
        self.forward_events_from(self.instance.executor)

    def __getstate__(self):
        state = super().__getstate__()
        state["modules"] = self.modules
        state["store"] = self.store
        state["stack"] = self.stack
        state["advice"] = self.advice
        state["constraints"] = self.constraints
        state["instantiated"] = self.instantiated
        state["module_names"] = self.module_names
        state["default_module"] = self.default_module
        state["manual_exports"] = self.manual_exports
        return state

    def __setstate__(self, state):
        self.modules = state["modules"]
        self.store = state["store"]
        self.stack = state["stack"]
        self.advice = state["advice"]
        self.constraints = state["constraints"]
        self.instantiated = state["instantiated"]
        self.module_names = state["module_names"]
        self.default_module = state["default_module"]
        self.manual_exports = state["manual_exports"]
        self.forward_events_from(self.stack)
        self.forward_events_from(self.instance)
        self.forward_events_from(self.instance.executor)
        for mem in self.store.mems:
            self.forward_events_from(mem)
        super().__setstate__(state)

    @property
    def instance(self) -> ModuleInstance:
        """
        :return: the ModuleInstance for the first module registered
        """
        return self.modules[self.module_names[self.default_module]][1]

    @property
    def module(self) -> Module:
        """
        :return: The first module registered
        """
        return self.modules[self.module_names[self.default_module]][0]

    def register_module(self, name, filename_or_alias):
        """
        Provide an explicit path to a WASM module so the importer will know where to find it

        :param name: Module name to register the module under
        :param filename_or_alias: Name of the .wasm file that module lives in
        :return:
        """
        if filename_or_alias in self.module_names:
            self.module_names[name] = self.module_names[filename_or_alias]
        if name not in self.module_names:
            self.modules.append((Module.load(filename_or_alias), ModuleInstance(self.constraints)))
            self.module_names[name] = len(self.modules) - 1
            self.instantiated.append(False)

    def set_env(
        self,
        exports: typing.Dict[
            str, typing.Union[ProtoFuncInst, TableInst, MemInst, GlobalInst, typing.Callable]
        ],
        mod_name="env",
    ):
        """
        Manually insert exports into the global environment

        :param exports: Dict mapping names to functions/tables/globals/memories
        :param mod_name: The name of the module these exports should fall under
        """
        self.manual_exports.setdefault(mod_name, {}).update(exports)

    def import_module(self, module_name, exec_start, stub_missing):
        """
        Collect all of the imports for the given module and instantiate it

        :param module_name: module to import
        :param exec_start: whether to run the start functions automatically
        :param stub_missing: whether to replace missing imports with stubs
        :return: None
        """
        search_paths = {"."}
        # If the module isn't registered, look for it on the filesystem
        if module_name not in self.module_names:
            logger.debug("Module %s was not provided, attempting to load from disk", module_name)
            for pth in search_paths:
                possible_path = os.path.join(pth, module_name + ".wasm")
                if os.path.isfile(possible_path):
                    self.register_module(module_name, possible_path)
                    break
            else:
                raise RuntimeError("Missing imported module: " + module_name)

        if self.instantiated[self.module_names[module_name]]:
            return

        # Get the module and the instance from the world
        module, instance = self.modules[self.module_names[module_name]]

        imports = self.get_module_imports(module, exec_start, stub_missing)

        instance.instantiate(self.store, module, imports, exec_start)
        self.instantiated[self.module_names[module_name]] = True
        logger.info("Imported %s", module_name)

    def _get_export_addr(
        self, export_name, mod_name=None
    ) -> typing.Optional[typing.Union[FuncAddr, TableAddr, MemAddr, GlobalAddr]]:
        """
        Gets the address in the store of a given export

        :param export_name: Name of the export to look for
        :param mod_name: Name of the module the export lives in
        :return: The address of the export
        """
        try:
            if mod_name in self.module_names:  # TODO - handle mod_name.export_name
                return self.modules[self.module_names[mod_name]][1].get_export_address(export_name)
        except MissingExportException as exc:
            logger.error("Couldn't find export %s.%s", mod_name, exc.name)
        return None

    def get_export(
        self, export_name, mod_name=None
    ) -> typing.Optional[
        typing.Union[ProtoFuncInst, TableInst, MemInst, GlobalInst, typing.Callable]
    ]:
        """
        Gets the export _instance_ for a given export & module name
        (basically just dereferences _get_export_addr into the store)

        :param export_name: Name of the export to look for
        :param mod_name: Name of the module the export lives in
        :return: The export itself
        """
        mod_name = self.default_module if not mod_name else mod_name
        if mod_name in self.manual_exports:
            if export_name in self.manual_exports[mod_name]:
                return self.manual_exports[mod_name][export_name]
        addr = self._get_export_addr(export_name, mod_name)
        if addr is not None:
            if isinstance(addr, FuncAddr):
                return self.store.funcs[addr]
            if isinstance(addr, TableAddr):
                return self.store.funcs[addr]
            if isinstance(addr, MemAddr):
                return self.store.mems[addr]
            if isinstance(addr, GlobalAddr):
                return self.store.globals[addr]

        return None

    def get_module_imports(self, module, exec_start, stub_missing) -> typing.List[ExternVal]:
        """
        Builds the list of imports that should be passed to the given module upon instantiation

        :param module: The module to find the imports for
        :param exec_start: Whether to execute the start function of the module
        :param stub_missing: Whether to replace missing imports with stubs (TODO: symbolicate)
        :return: List of addresses for the imports within the store
        """
        imports: typing.List[ExternVal] = []
        for i in module.imports:
            logger.info("Importing %s:%s", i.module, i.name)
            if i.module not in self.module_names:  # If the module isn't registered
                if i.module not in self.manual_exports:  # or manually provided
                    # Attempt to load it from the filesystem
                    self.import_module(
                        i.module, exec_start, stub_missing
                    )  # TODO - prevent circular imports
            # If it's registered, but hasn't been imported yet, import it
            elif not self.instantiated[self.module_names[i.module]]:
                self.import_module(i.module, exec_start, stub_missing)
            imported_version = self._get_export_addr(i.name, i.module)
            if imported_version is None:
                # check for manually provided version
                # Possibly overwriting imported_version from an address to an instance here
                imported_version = self.get_export(i.name, i.module)  # type: ignore
                if imported_version is None and not stub_missing:
                    raise RuntimeError(f"Could not find import {i.module}:{i.name}")

            if isinstance(imported_version, (FuncAddr, TableAddr, MemAddr, GlobalAddr)):
                imports.append(imported_version)  # TODO - Import type matching
            else:
                if isinstance(i.desc, TypeIdx):  # Imported function
                    func_type = module.types[i.desc]
                    if i.module == "env" and imported_version:
                        if callable(imported_version):
                            logger.info(
                                "Auto-converting callable %s into HostFunc of appropriate type",
                                i.name,
                            )
                            imported_version = HostFunc(func_type, imported_version)
                    self.store.funcs.append(
                        imported_version
                        if imported_version
                        else HostFunc(
                            # mypy doesn't like that we're using a partial function instead of a proper function
                            func_type,
                            partial(stub, len(func_type.result_types)),  # type: ignore
                        )
                    )
                    addr = FuncAddr(len(self.store.funcs) - 1)
                    imports.append(addr)
                    self.instance.function_names[addr] = f"{i.module}.{i.name}"

                elif isinstance(i.desc, TableType):
                    self.store.tables.append(
                        imported_version
                        if imported_version
                        else TableInst([None] * i.desc.limits.min, i.desc.limits.max)
                    )
                    imports.append(TableAddr(len(self.store.tables) - 1))
                # Create an empty memory of the correct size and provide it as an import
                elif isinstance(i.desc, MemoryType):
                    self.store.mems.append(
                        imported_version
                        if imported_version
                        else MemInst([0x0] * i.desc.min * 64 * 1024, i.desc.max)
                    )
                    imports.append(MemAddr(len(self.store.mems) - 1))
                # Create a global and set its value to 0.
                elif isinstance(i.desc, GlobalType):
                    self.store.globals.append(
                        imported_version
                        if imported_version
                        else GlobalInst(i.desc.value(0), i.desc.mut)
                    )
                    imports.append(GlobalAddr(len(self.store.globals) - 1))
                else:
                    raise RuntimeError(f"Don't know how to handle imports of type {type(i.desc)}")

        return imports

    def instantiate(
        self,
        env_import_dict: typing.Dict[
            str, typing.Union[ProtoFuncInst, TableInst, MemInst, GlobalInst, typing.Callable]
        ],
        supplemental_env: typing.Dict[
            str,
            typing.Dict[
                str, typing.Union[ProtoFuncInst, TableInst, MemInst, GlobalInst, typing.Callable]
            ],
        ] = {},
        exec_start=False,
        stub_missing=True,
    ):
        """
        Prepares the underlying ModuleInstance for execution. Calls import_module under the hood, so this is probably
        the only import-y function you ever need to call externally.

        TODO: stubbed imports should be symbolic

        :param env_import_dict: Dict mapping strings to functions. Functions should accept the current ConstraintSet as the first argument.
        :param supplemental_env: Maps strings w/ module names to environment dicts using the same format as env_import_dict
        :param exec_start: Whether or not to automatically execute the `start` function, if it is set.
        :param stub_missing: Whether or not to replace missing imports with empty stubs
        :return: None
        """
        self.set_env(env_import_dict)
        for k in supplemental_env:
            self.set_env(supplemental_env[k], k)
        self.import_module(self.default_module, exec_start, stub_missing)
        for mem in self.store.mems:
            self.forward_events_from(mem)

    def invoke(self, name="main", argv=[], module=None):
        """
        Sets up the WASMWorld to run the function specified by `name` when `ManticoreWASM.run` is called

        :param name: Name of the function to invoke
        :param argv: List of arguments to pass to the function. Should typically be I32, I64, F32, or F64
        :param module: name of a module to call the function in (if not the default module)
        :return: None
        """
        module = self.default_module if module is None else module
        instance = self.modules[self.module_names[module]][1]
        instance.invoke_by_name(name, self.stack, self.store, list(argv))

    def exec_for_test(self, funcname, module=None):
        """
        Helper method that simulates the evaluation loop without creating workers or states, forking, or concretizing
        symbolic values.
        Only used for concrete unit testing.

        :param funcname: The name of the function to test
        :param module: The name of the module to test the function in (if not the default module)
        :return: The top n items from the stack where n is the expected number of return values from the function
        """
        module = self.default_module if module is None else module
        instance = self.modules[self.module_names[module]][1]

        # Grab the appropriate number of return values for the function being invoked
        rets = 0
        for export in instance.exports:
            if export.name == funcname and isinstance(export.value, FuncAddr):
                rets = len(self.store.funcs[export.value].type.result_types)

        # Call exec_instruction until it returns false or throws an error
        try:
            while self._exec_instruction(instance):
                pass
            # Return the top `rets` values from the stack
            return [self.stack.pop() for _i in range(rets)]
        except (Trap, NotImplementedError) as e:  # Reset the internals if we have any problems
            instance.reset_internal()
            self.stack = Stack()
            raise e

    def execute(self, current_state):
        """
        Tells the underlying ModuleInstance to execute a single WASM instruction. Raises TerminateState if there are
        no more instructions to execute, or if the instruction raises a Trap.
        """
        if not self.instantiated:
            raise RuntimeError("Trying to execute before instantiation!")
        try:
            if not self._exec_instruction(self.instance, current_state):
                raise TerminateState(f"Execution returned {self.stack.peek()}")
        except Trap as e:
            raise TerminateState(f"Execution raised Trap: {str(e)}")

    def _exec_instruction(self, instance, current_state=None):
        """
        Executes a single instruction on the instance and clears the advice.
        """
        try:
            return instance.exec_instruction(self.store, self.stack, self.advice, current_state)
        finally:
            self.advice = None



import sys

from functools import reduce
import re
import logging

from ..core.plugin import Plugin
from ..core.smtlib import Operators, to_constant
from ..utils.enums import StateLists
import pyevmasm as EVMAsm

logger = logging.getLogger(__name__)


class FilterFunctions(Plugin):
    def __init__(
        self, regexp=r".*", mutability="both", depth="both", fallback=False, include=True, **kwargs
    ):
        """
        Constrain input based on function metadata. Include or avoid functions
        selected by the specified criteria.

        Examples:
        #Do not explore any human transactions that end up calling a constant function
        no_human_constant = FilterFunctions(depth='human', mutability='constant', include=False)

        #At human tx depth only accept synthetic check functions
        only_tests = FilterFunctions(regexp=r'mcore_.*', depth='human', include=False)

        :param regexp: a regular expression over the name of the function '.*' will match all functions
        :param mutability: mutable, constant or both will match functions declared in the abi to be of such class
        :param depth: match functions in internal transactions, in human initiated transactions or in both types
        :param fallback: if True include the fallback function. Hash will be 00000000 for it
        :param include: if False exclude the selected functions, if True include them
        """
        super().__init__(**kwargs)
        depth = depth.lower()
        if depth not in ("human", "internal", "both"):
            raise ValueError
        mutability = mutability.lower()
        if mutability not in ("mutable", "constant", "both"):
            raise ValueError

        # fixme better names for member variables
        self._regexp = regexp
        self._mutability = mutability
        self._depth = depth
        self._fallback = fallback
        self._include = include

    def will_open_transaction_callback(self, state, tx):
        world = state.platform
        tx_cnt = len(world.all_transactions)
        # Constrain input only once per tx, per plugin
        if state.context.get("constrained%d" % id(self), 0) != tx_cnt:
            state.context["constrained%d" % id(self)] = tx_cnt

            if self._depth == "human" and not tx.is_human:
                return
            if self._depth == "internal" and tx.is_human:
                return

            # Get metadata if any for the target address of current tx
            md = self.manticore.get_metadata(tx.address)
            if md is None:
                return
            # Let's compile  the list of interesting hashes
            selected_functions = []

            for func_hsh in md.function_selectors:
                abi = md.get_abi(func_hsh)
                if abi["type"] == "fallback":
                    continue
                if self._mutability == "constant" and not abi.get("constant", False):
                    continue
                if self._mutability == "mutable" and abi.get("constant", False):
                    continue
                if not re.match(self._regexp, abi["name"]):
                    continue
                selected_functions.append(func_hsh)

            if self._fallback and md.has_non_default_fallback_function:
                selected_functions.append(md.fallback_function_selector)

            if self._include:
                if not selected_functions:
                    logger.warning("No functions selected, adding False to path constraint.")
                # constrain the input so it can take only the interesting values
                constraint = reduce(
                    Operators.OR, (tx.data[:4] == x for x in selected_functions), False
                )
                state.constrain(constraint)
            else:
                # Avoid all not selected hashes
                constraint = True
                for func_hsh in md.function_selectors:
                    if func_hsh in selected_functions:
                        constraint = Operators.AND(tx.data[:4] != func_hsh, constraint)
                state.constrain(constraint)


class LoopDepthLimiter(Plugin):
    """This just aborts explorations that are too deep"""

    def __init__(self, loop_count_threshold=5, **kwargs):
        super().__init__(**kwargs)
        self.loop_count_threshold = loop_count_threshold

    def will_run_callback(self, *args):
        with self.manticore.locked_context("seen_rep", dict) as reps:
            reps.clear()

    def will_execute_instruction_callback(self, state, pc, insn):
        world = state.platform
        with self.manticore.locked_context("seen_rep", dict) as reps:
            item = (
                world.current_transaction.sort == "CREATE",
                world.current_transaction.address,
                pc,
            )
            if item not in reps:
                reps[item] = 0
            reps[item] += 1
            if reps[item] > self.loop_count_threshold:
                state.abandon()


class VerboseTrace(Plugin):
    """
    Generates a verbose trace of EVM execution and saves in workspace into `state<id>.trace`.

    Example output can be seen in test_eth_plugins.
    """

    def will_evm_execute_instruction_callback(self, state, instruction, arguments):
        current_vm = state.platform.current_vm
        state.context.setdefault("str_trace", []).append(str(current_vm))

    def generate_testcase(self, state, testcase, message):
        trace = state.context.get("str_trace", [])

        with testcase.open_stream("verbose_trace") as vt:
            for t in trace:
                vt.write(t + "\n")


class VerboseTraceStdout(Plugin):
    """
    Same as VerboseTrace but prints to stdout. Note that you should use it only if Manticore
    is run with procs=1 as otherwise, the output will be clobbered.
    """

    def will_evm_execute_instruction_callback(self, state, instruction, arguments):
        print(state.platform.current_vm)


class KeepOnlyIfStorageChanges(Plugin):
    """This plugin discards all transactions that results in states where
    the underlying EVM storage did not change or in other words,
    there were no writes to it.

    This allows to speed-up EVM engine exploration as we don't
    explore states that have the same storage (contract data).

    However, keep in mind that if the (contract) code relies on
    account balance and the balance is not a symbolic value
    it might be that a certain state will not be covered by the
    execution when this plugin is used.
    """

    def did_open_transaction_callback(self, state, tx, *args):
        """We need a stack. Each tx (internal or not) starts with a "False" flag
        denoting that it did not write anything to the storage
        """
        state.context["written"].append(False)

    def did_close_transaction_callback(self, state, tx, *args):
        """When a tx (internal or not) is closed a value is popped out from the
        flag stack. Depending on the result if the storage is not rolled back the
        next flag in the stack is updated. Not that if the a tx is reverted the
        changes it may have done on the storage will not affect the final result.

        """
        flag = state.context["written"].pop()
        if tx.result in {"RETURN", "STOP"}:
            code_written = (tx.result == "RETURN") and (tx.sort == "CREATE")
            flag = flag or code_written
            # As the ether balance of any account can be manipulated beforehand
            # it does not matter if a state can affect the balances or not.
            # The same reachability should be obtained as the account original
            # balances must be symbolic and free-ish
            if not flag:
                ether_sent = state.can_be_true(tx.value != 0)
                flag = flag or ether_sent
            state.context["written"][-1] = state.context["written"][-1] or flag

    def did_evm_write_storage_callback(self, state, *args):
        """Turn on the corresponding flag is the storage has been modified.
        Note: subject to change if the current transaction is reverted"""
        state.context["written"][-1] = True

    def will_run_callback(self, *args):
        """Initialize the flag stack at each human tx/run()"""
        for st in self.manticore.ready_states:
            st.context["written"] = [False]

    def did_run_callback(self):
        """When  human tx/run just ended remove the states that have not changed
        the storage"""
        with self.manticore.locked_context("ethereum.saved_states", list) as saved_states:
            # Normally the ready states are consumed and forked, eth save the
            # states that finished ok in a special context list. This list will
            # compose the ready states for the next human transaction.
            # The actual "ready_states" list at this point contain the states
            # that have not finished the previous TX due to a timeout. Those will
            # be ignored.
            for state_id in list(saved_states):
                st = self.manticore._load(state_id)
                if not st.context["written"][-1]:
                    if st.id in self.manticore._ready_states:
                        self._publish(
                            "will_transition_state",
                            state_id,
                            StateLists.ready,
                            StateLists.terminated,
                        )
                        self.manticore._ready_states.remove(st.id)
                        self.manticore._terminated_states.append(st.id)
                        self._publish(
                            "did_transition_state",
                            state_id,
                            StateLists.ready,
                            StateLists.terminated,
                        )
                    saved_states.remove(st.id)

    def generate_testcase(self, state, testcase, message):
        with testcase.open_stream("summary") as stream:
            if not state.context.get("written", (False,))[-1]:
                stream.write(
                    "State was removed from ready list because the last tx did not write to the storage"
                )


class SkipRevertBasicBlocks(Plugin):
    def _is_revert_bb(self, state, pc):
        world = state.platform

        def read_code(_pc=None):
            while True:
                yield to_constant(world.current_vm.read_code(_pc)[0])
                _pc += 1

        for inst in EVMAsm.disassemble_all(read_code(pc), pc):
            if inst.name == "REVERT":
                return True
            if inst.is_terminator:
                return False

    def will_evm_execute_instruction_callback(self, state, instruction, arguments):
        world = state.platform
        if state.platform.current_transaction.sort != "CREATE":
            if instruction.semantics == "JUMPI":

                # if the bb after the jumpi ends ina revert do not explore it.
                if self._is_revert_bb(state, world.current_vm.pc + instruction.size):
                    state.constrain(arguments[1] == True)

                # if the target of the jumpi ends up in a revert avoid it
                if self._is_revert_bb(state, arguments[0]):
                    state.constrain(arguments[1] == False)

                # This may have added an impossible constraint.



from .manticore import ManticoreWASM

from ..core.plugin import Profiler
from ..utils import config

consts = config.get_group("cli")
consts.add("profile", default=False, description="Enable worker profiling mode")
consts.add("target_func", default="main", description="WASM Function to execute")


def wasm_main(args, _logger):

    m = ManticoreWASM(
        args.argv[0],
        argv=args.argv[1:],
        env={},
        exec_start=True,
        workspace_url=args.workspace,
        policy=args.policy,
    )

    if consts.profile:
        profiler = Profiler()
        m.register_plugin(profiler)

    m.default_invoke(func_name=consts.target_func)

    with m.kill_timeout():
        m.run()

    m.finalize()

    return m



import inspect
import io
import logging
import struct
import types
from functools import wraps, partial
from itertools import islice

import unicorn

from .disasm import init_disassembler, Instruction
from ..memory import ConcretizeMemory, InvalidMemoryAccess, FileMap, AnonMap
from ..memory import LazySMemory, Memory
from ...core.smtlib import Operators, Constant, issymbolic, BitVec, Expression
from ...core.smtlib import visitors
from ...core.smtlib.solver import SelectedSolver
from ...utils.emulate import ConcreteUnicornEmulator
from ...utils.event import Eventful
from ...utils.fallback_emulator import UnicornEmulator

from capstone import CS_ARCH_ARM64, CS_ARCH_X86, CS_ARCH_ARM
from capstone.arm64 import ARM64_REG_ENDING
from capstone.x86 import X86_REG_ENDING
from capstone.arm import ARM_REG_ENDING

from typing import Any, Callable, Dict, Optional, Tuple

logger = logging.getLogger(__name__)
register_logger = logging.getLogger(f"{__name__}.registers")


def _sig_is_varargs(sig: inspect.Signature) -> bool:
    VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
    return any(p.kind == VAR_POSITIONAL for p in sig.parameters.values())


###################################################################################
# Exceptions


class CpuException(Exception):
    """Base cpu exception"""


class DecodeException(CpuException):
    """
    Raised when trying to decode an unknown or invalid instruction"""

    def __init__(self, pc, bytes):
        super().__init__("Error decoding instruction @ 0x{:x}".format(pc))
        self.pc = pc
        self.bytes = bytes


class InstructionNotImplementedError(CpuException):
    """
    Exception raised when you try to execute an instruction that is not yet
    implemented in the emulator. Add it to the Cpu-specific implementation.
    """


class InstructionEmulationError(CpuException):
    """
    Exception raised when failing to emulate an instruction outside of Manticore.
    """


class DivideByZeroError(CpuException):
    """A division by zero"""


class Interruption(CpuException):
    """A software interrupt."""

    def __init__(self, N):
        super().__init__("CPU Software Interruption %08x" % N)
        self.N = N


class Syscall(CpuException):
    """ """

    def __init__(self):
        super().__init__("CPU Syscall")


class ConcretizeRegister(CpuException):
    """
    Raised when a symbolic register needs to be concretized.
    """

    def __init__(
        self,
        cpu: "Cpu",
        reg_name: str,
        message: Optional[str] = None,
        policy: str = "MINMAX",
    ):
        self.message = message if message else f"Concretizing {reg_name}"

        self.cpu = cpu
        self.reg_name = reg_name
        self.policy = policy


class ConcretizeArgument(CpuException):
    """
    Raised when a symbolic argument needs to be concretized.
    """

    def __init__(self, cpu, argnum, policy="MINMAX"):
        self.message = f"Concretizing argument #{argnum}."
        self.cpu = cpu
        self.policy = policy
        self.argnum = argnum


SANE_SIZES = {8, 16, 32, 64, 80, 128, 256}


class Operand:
    """This class encapsulates how to access operands (regs/mem/immediates) for
    different CPUs
    """

    class MemSpec:
        """
        Auxiliary class wraps capstone operand 'mem' attribute. This will
        return register names instead of Ids
        """

        def __init__(self, parent):
            self.parent = parent

        segment = property(lambda self: self.parent._reg_name(self.parent.op.mem.segment))
        base = property(lambda self: self.parent._reg_name(self.parent.op.mem.base))
        index = property(lambda self: self.parent._reg_name(self.parent.op.mem.index))
        scale = property(lambda self: self.parent.op.mem.scale)
        disp = property(lambda self: self.parent.op.mem.disp)

    def __init__(self, cpu: "Cpu", op):
        """
        This encapsulates the arch-independent way to access instruction
        operands and immediates based on the disassembler operand descriptor in
        use. This class knows how to browse an operand and get its details.

        It also knows how to access the specific Cpu to get the actual values
        from memory and registers.

        :param Cpu cpu: A Cpu instance
        :param Operand op: An wrapped Instruction Operand
        :type op: X86Op or ArmOp
        """
        assert isinstance(cpu, Cpu)
        self.cpu = cpu
        self.op = op
        self.mem = Operand.MemSpec(self)

    def _reg_name(self, reg_id: int):
        """
        Translates a register ID from the disassembler object into the
        register name based on manticore's alias in the register file

        :param reg_id: Register ID
        """
        # XXX: Support other architectures.
        if (
            (self.cpu.arch == CS_ARCH_ARM64 and reg_id >= ARM64_REG_ENDING)
            or (self.cpu.arch == CS_ARCH_X86 and reg_id >= X86_REG_ENDING)
            or (self.cpu.arch == CS_ARCH_ARM and reg_id >= ARM_REG_ENDING)
        ):
            logger.warning("Trying to get register name for a non-register")
            return None
        cs_reg_name = self.cpu.instruction.reg_name(reg_id)
        if cs_reg_name is None or cs_reg_name.lower() == "(invalid)":
            return None
        return self.cpu._regfile._alias(cs_reg_name.upper())

    def __getattr__(self, name):
        return getattr(self.op, name)

    @property
    def type(self):
        """This property encapsulates the operand type.
        It may be one of the following:
            register
            memory
            immediate
        """
        raise NotImplementedError

    @property
    def size(self):
        """Return bit size of operand"""
        raise NotImplementedError

    @property
    def reg(self):
        return self._reg_name(self.op.reg)

    def address(self):
        """On a memory operand it returns the effective address"""
        raise NotImplementedError

    def read(self):
        """It reads the operand value from the registers or memory"""
        raise NotImplementedError

    def write(self, value):
        """It writes the value of specific type to the registers or memory"""
        raise NotImplementedError


#  Basic register file structure not actually need to abstract as it's used
#  only from the cpu implementation


class RegisterFile:
    def __init__(self, aliases=None):
        # dict mapping from alias register name ('PC') to actual register
        # name ('RIP')
        self._aliases = aliases if aliases is not None else {}
        self._registers = {}

    def _alias(self, register):
        """
        Get register canonical alias. ex. PC->RIP or PC->R15

        :param str register: The register name
        """
        return self._aliases.get(register, register)

    def write(self, register, value):
        """
        Write value to the specified register

        :param str register: a register id. Must be listed on all_registers
        :param value: a value of the expected type
        :type value: int or long or Expression
        :return: the value actually written to the register
        """
        raise NotImplementedError

    def read(self, register):
        """
        Read value from specified register

        :param str register: a register name. Must be listed on all_registers
        :return: the register value
        """
        raise NotImplementedError

    @property
    def all_registers(self):
        """Lists all possible register names (Including aliases)"""
        return tuple(self._aliases)

    @property
    def canonical_registers(self):
        """List the minimal most beautiful set of registers needed"""
        raise NotImplementedError

    def __contains__(self, register):
        """
        Check for register validity

        :param register: a register name
        """
        return self._alias(register) in self.all_registers

    def __copy__(self) -> "RegisterFile":
        """Custom shallow copy to create a snapshot of the register state.
        Should be used as read-only"""
        ...


class Abi:
    """
    Represents the ability to extract arguments from the environment and write
    back a result.

    Used for function call and system call models.
    """

    def __init__(self, cpu: "Cpu"):
        """
        :param CPU to initialize with
        """
        self._cpu = cpu

    def get_arguments(self):
        """
        Extract model arguments conforming to `convention`. Produces an iterable
        of argument descriptors following the calling convention. A descriptor
        is either a string describing a register, or an address (concrete or
        symbolic).

        :return: iterable returning syscall arguments.
        :rtype: iterable
        """
        raise NotImplementedError

    def get_result_reg(self) -> str:
        """
        Extract the location a return value will be written to. Produces
        a string describing a register where the return value is written to.
        :return: return register name
        :rtype: string
        """
        raise NotImplementedError

    def write_result(self, result):
        """
        Write the result of a model back to the environment.

        :param result: result of the model implementation
        """
        raise NotImplementedError

    def ret(self):
        """
        Handle the "ret" semantics of the ABI, i.e. reclaiming stack space,
        popping PC, etc.

        A null operation by default.
        """
        return

    def values_from(self, base):
        """
        A reusable generator for increasing pointer-sized values from an address
        (usually the stack).
        """
        word_bytes = self._cpu.address_bit_size // 8
        while True:
            yield base
            base += word_bytes

    def get_argument_values(self, model: Callable, prefix_args: Tuple) -> Tuple:
        """
        Extract arguments for model from the environment and return as a tuple that
        is ready to be passed to the model.

        :param model: Python model of the function
        :param prefix_args: Parameters to pass to model before actual ones
        :return: Arguments to be passed to the model
        """
        if type(model) is partial:
            # mypy issue with partial types https://github.com/python/mypy/issues/1484
            model = model.args[0]  # type: ignore
        sig = inspect.signature(model)
        if _sig_is_varargs(sig):
            model_name = getattr(model, "__qualname__", "<no name>")
            logger.warning("ABI: %s: a vararg model must be a unary function.", model_name)

        nargs = len(sig.parameters) - len(prefix_args)

        def resolve_argument(arg):
            if isinstance(arg, str):
                return self._cpu.read_register(arg)
            else:
                return self._cpu.read_int(arg)

        # Create a stream of resolved arguments from argument descriptors
        descriptors = self.get_arguments()
        argument_iter = map(resolve_argument, descriptors)

        from ..models import isvariadic  # prevent circular imports

        if isvariadic(model):
            return prefix_args + (argument_iter,)
        else:
            return prefix_args + tuple(islice(argument_iter, nargs))

    def invoke(self, model, prefix_args=None):
        """
        Invoke a callable `model` as if it was a native function. If
        :func:`~manticore.models.isvariadic` returns true for `model`, `model` receives a single
        argument that is a generator for function arguments. Pass a tuple of
        arguments for `prefix_args` you'd like to precede the actual
        arguments.

        :param callable model: Python model of the function
        :param tuple prefix_args: Parameters to pass to model before actual ones
        :return: The result of calling `model`
        """
        prefix_args = prefix_args or ()

        arguments = self.get_argument_values(model, prefix_args)

        try:
            result = model(*arguments)
        except ConcretizeArgument as e:
            assert e.argnum >= len(prefix_args), "Can't concretize a constant arg"
            idx = e.argnum - len(prefix_args)

            # Arguments were lazily computed in case of variadic, so recompute here
            descriptors = self.get_arguments()
            src = next(islice(descriptors, idx, idx + 1))

            msg = "Concretizing due to model invocation"
            if isinstance(src, str):
                raise ConcretizeRegister(self._cpu, src, msg)
            else:
                raise ConcretizeMemory(self._cpu.memory, src, self._cpu.address_bit_size, msg)
        else:
            if result is not None:
                self.write_result(result)

            self.ret()

        return result


platform_logger = logging.getLogger("manticore.platforms.platform")


def unsigned_hexlify(i: Any) -> Any:
    if type(i) is int:
        if i < 0:
            return hex((1 << 64) + i)
        return hex(i)
    return i


class SyscallAbi(Abi):
    """
    A system-call specific ABI.

    Captures model arguments and return values for centralized logging.
    """

    def syscall_number(self):
        """
        Extract the index of the invoked syscall.

        :return: int
        """
        raise NotImplementedError

    def get_argument_values(self, model, prefix_args):
        self._last_arguments = super().get_argument_values(model, prefix_args)
        return self._last_arguments

    def invoke(self, model, prefix_args=None):
        # invoke() will call get_argument_values()
        self._last_arguments = ()

        if type(model) is partial:
            self._cpu._publish("will_execute_syscall", model.args[0])
        else:
            self._cpu._publish("will_execute_syscall", model)
        ret = super().invoke(model, prefix_args)
        if type(model) is partial:
            model = model.args[0]
        self._cpu._publish(
            "did_execute_syscall",
            model.__func__.__name__ if isinstance(model, types.MethodType) else model.__name__,
            self._last_arguments,
            ret,
        )

        if platform_logger.isEnabledFor(logging.DEBUG):
            # Try to expand strings up to max_arg_expansion
            max_arg_expansion = 32
            # Add a hex representation to return if greater than min_hex_expansion
            min_hex_expansion = 0x80

            args = []
            for arg in self._last_arguments:
                arg_s = (
                    unsigned_hexlify(arg)
                    if not issymbolic(arg) and abs(arg) > min_hex_expansion
                    else f"{arg}"
                )
                if self._cpu.memory.access_ok(arg, "r") and model.__func__.__name__ not in {
                    "sys_mprotect",
                    "sys_mmap",
                }:
                    try:
                        s = self._cpu.read_string(arg, max_arg_expansion)
                        s = s.rstrip().replace("\n", "\\n") if s else s
                        arg_s = f'"{s}"' if s else arg_s
                    except Exception:
                        pass
                args.append(arg_s)

            args_s = ", ".join(args)
            ret_s = f"{unsigned_hexlify(ret)}" if abs(ret) > min_hex_expansion else f"{ret}"

            platform_logger.debug("%s(%s) = %s", model.__func__.__name__, args_s, ret_s)


############################################################################
# Abstract cpu encapsulating common cpu methods used by platforms and executor.


class Cpu(Eventful):
    """
    Base class for all Cpu architectures. Functionality common to all
    architectures (and expected from users of a Cpu) should be here. Commonly
    used by platforms and py:class:manticore.core.Executor

    The following attributes need to be defined in any derived class

    - arch
    - mode
    - max_instr_width
    - address_bit_size
    - pc_alias
    - stack_alias
    """

    _published_events = {
        "write_register",
        "read_register",
        "write_memory",
        "read_memory",
        "decode_instruction",
        "execute_instruction",
        "invoke_syscall",
        "set_descriptor",
        "map_memory",
        "protect_memory",
        "unmap_memory",
        "execute_syscall",
        "solve",
    }

    def __init__(self, regfile: RegisterFile, memory: Memory, **kwargs):
        assert isinstance(regfile, RegisterFile)
        self._disasm = kwargs.pop("disasm", "capstone")
        super().__init__(**kwargs)
        self._regfile = regfile
        self._memory = memory
        self._instruction_cache: Dict[int, Instruction] = {}
        self._icount = 0
        # _last_pc represents the last PC that was going to be executed, but it
        # might not have been due to user hooks, exceptions, etc. You probably
        # want last_executed_pc() or last_executed_insn()
        self._last_pc = None
        # _last_executed_pc represents the last PC that was executed and
        # affected the state of the program
        self._last_executed_pc = None
        self._concrete = kwargs.pop("concrete", False)
        self.emu = None
        self._break_unicorn_at: Optional[int] = None
        self._delayed_event = False
        if not hasattr(self, "disasm"):
            self.disasm = init_disassembler(self._disasm, self.arch, self.mode)
        # Ensure that regfile created STACK/PC aliases
        assert "STACK" in self._regfile
        assert "PC" in self._regfile

    def __getstate__(self):
        state = super().__getstate__()
        state["regfile"] = self._regfile
        state["memory"] = self._memory
        state["icount"] = self._icount
        state["last_pc"] = self._last_pc
        state["last_executed_pc"] = self._last_executed_pc
        state["disassembler"] = self._disasm
        state["concrete"] = self._concrete
        state["break_unicorn_at"] = self._break_unicorn_at
        state["delayed_event"] = self._delayed_event
        return state

    def __setstate__(self, state):
        Cpu.__init__(
            self,
            state["regfile"],
            state["memory"],
            disasm=state["disassembler"],
            concrete=state["concrete"],
        )
        self._icount = state["icount"]
        self._last_pc = state["last_pc"]
        self._last_executed_pc = state["last_executed_pc"]
        self._disasm = state["disassembler"]
        self._concrete = state["concrete"]
        self._break_unicorn_at = state["break_unicorn_at"]
        self._delayed_event = state["delayed_event"]
        super().__setstate__(state)

    @property
    def icount(self):
        return self._icount

    @property
    def last_executed_pc(self) -> Optional[int]:
        """The last PC that was executed."""
        return self._last_executed_pc

    @property
    def last_executed_insn(self) -> Optional[Instruction]:
        """The last instruction that was executed."""
        if not self.last_executed_pc:
            return None
        return self.decode_instruction(self.last_executed_pc)

    ##############################
    # Register access
    @property
    def regfile(self):
        """The RegisterFile of this cpu"""
        return self._regfile

    @property
    def all_registers(self):
        """
        Returns all register names for this CPU. Any register returned can be
        accessed via a `cpu.REG` convenience interface (e.g. `cpu.EAX`) for both
        reading and writing.

        :return: valid register names
        :rtype: tuple[str]
        """
        return self._regfile.all_registers

    @property
    def canonical_registers(self):
        """
        Returns the list of all register names  for this CPU.

        :rtype: tuple
        :return: the list of register names for this CPU.
        """
        return self._regfile.canonical_registers

    def write_register(self, register, value):
        """
        Dynamic interface for writing cpu registers

        :param str register: register name (as listed in `self.all_registers`)
        :param value: register value
        :type value: int or long or Expression
        """
        self._publish("will_write_register", register, value)
        value = self._regfile.write(register, value)
        self._publish("did_write_register", register, value)
        return value

    def read_register(self, register):
        """
        Dynamic interface for reading cpu registers

        :param str register: register name (as listed in `self.all_registers`)
        :return: register value
        :rtype: int or long or Expression
        """
        self._publish("will_read_register", register)
        value = self._regfile.read(register)
        self._publish("did_read_register", register, value)
        return value

    # Pythonic access to registers and aliases
    def __getattr__(self, name):
        """
        A Pythonic version of read_register

        :param str name: Name of the register
        """
        if name != "_regfile":
            if name in self._regfile:
                return self.read_register(name)
        raise AttributeError(name)

    def __setattr__(self, name, value):
        """
        A Pythonic version of write_register

        :param str name: Name of the register to set
        :param value: The value to set the register to
        :type param: int or long or Expression
        """
        try:
            if name in self._regfile:
                return self.write_register(name, value)
            object.__setattr__(self, name, value)
        except AttributeError:
            object.__setattr__(self, name, value)

    def emulate_until(self, target: int):
        """
        Tells the CPU to set up a concrete unicorn emulator and use it to execute instructions
        until target is reached.

        :param target: Where Unicorn should hand control back to Manticore. Set to 0 for all instructions.
        """
        self._concrete = True
        self._break_unicorn_at = target
        if self.emu:
            self.emu.write_backs_disabled = False
            self.emu.load_state_from_manticore()
            self.emu._stop_at = target

    #############################
    # Memory access
    @property
    def memory(self) -> Memory:
        return self._memory

    def write_int(self, where, expression, size=None, force=False):
        """
        Writes int to memory

        :param int where: address to write to
        :param expr: value to write
        :type expr: int or BitVec
        :param size: bit size of `expr`
        :param force: whether to ignore memory permissions
        """
        if size is None:
            size = self.address_bit_size
        assert size in SANE_SIZES
        self._publish("will_write_memory", where, expression, size)

        data = [
            Operators.CHR(Operators.EXTRACT(expression, offset, 8)) for offset in range(0, size, 8)
        ]
        self._memory.write(where, data, force)

        self._publish("did_write_memory", where, expression, size)

    def _raw_read(self, where: int, size: int = 1, force: bool = False) -> bytes:
        """
        Selects bytes from memory. Attempts to do so faster than via read_bytes.

        :param where: address to read from
        :param size: number of bytes to read
        :param force: whether to ignore memory permissions
        :return: the bytes in memory
        """
        map = self.memory.map_containing(where)
        start = map._get_offset(where)
        if isinstance(map, FileMap):
            end = map._get_offset(where + size)

            if end > map._mapped_size:
                logger.warning(
                    f"Missing {end - map._mapped_size} bytes at the end of {map._filename}"
                )

            raw_data = map._data[map._get_offset(where) : min(end, map._mapped_size)]
            if len(raw_data) < end:
                raw_data += b"\x00" * (end - len(raw_data))

            data = b""
            for offset in sorted(map._overlay.keys()):
                data += raw_data[len(data) : offset]
                data += map._overlay[offset]
            data += raw_data[len(data) :]

        elif isinstance(map, AnonMap):
            data = bytes(map._data[start : start + size])
        else:
            data = b"".join(self.memory.read(where, size, force=force))
        assert len(data) == size, "Raw read resulted in wrong data read which should never happen"
        return data

    def read_int(self, where: int, size: int = None, force: bool = False, publish: bool = True):
        """
        Reads int from memory

        :param where: address to read from
        :param size: number of bits to read
        :param force: whether to ignore memory permissions
        :param publish: whether to publish an event
        :return: the value read
        """
        if size is None:
            size = self.address_bit_size
        assert size in SANE_SIZES
        if publish:
            self._publish("will_read_memory", where, size)

        data = self._memory.read(where, size // 8, force)
        assert (8 * len(data)) == size
        value = Operators.CONCAT(size, *map(Operators.ORD, reversed(data)))

        if publish:
            self._publish("did_read_memory", where, value, size)
        return value

    def write_bytes(self, where: int, data, force: bool = False) -> None:
        """
        Write a concrete or symbolic (or mixed) buffer to memory

        :param where: address to write to
        :param data: data to write
        :param force: whether to ignore memory permissions
        """

        mp = self.memory.map_containing(where)
        # TODO (ehennenfent) - fast write can have some yet-unstudied unintended side effects.
        # At the very least, using it in non-concrete mode will break the symbolic strcmp/strlen models. The 1024 byte
        # minimum is intended to minimize the potential effects of this by ensuring that if there _are_ any other
        # issues, they'll only crop up when we're doing very large writes, which are fairly uncommon.
        if (
            isinstance(mp, AnonMap)
            and isinstance(data, (str, bytes))
            and (mp.end - mp.start + 1) >= len(data) >= 1024
            and not issymbolic(data)
            and self._concrete
        ):
            logger.debug("Using fast write")
            offset = mp._get_offset(where)
            if isinstance(data, str):
                data = bytes(data.encode("utf-8"))
            self._publish("will_write_memory", where, data, 8 * len(data))
            mp._data[offset : offset + len(data)] = data
            self._publish("did_write_memory", where, data, 8 * len(data))
        else:
            for i in range(len(data)):
                self.write_int(where + i, Operators.ORD(data[i]), 8, force)

    def read_bytes(self, where: int, size: int, force: bool = False, publish: bool = True):
        """
        Read from memory.

        :param where: address to read data from
        :param size: number of bytes
        :param force: whether to ignore memory permissions
        :param publish: whether to publish events
        :return: data
        """
        result = []
        for i in range(size):
            result.append(Operators.CHR(self.read_int(where + i, 8, force, publish=publish)))
        return result

    def write_string(
        self, where: int, string: str, max_length: Optional[int] = None, force: bool = False
    ) -> None:
        """
        Writes a string to memory, appending a NULL-terminator at the end.

        :param where: Address to write the string to
        :param string: The string to write to memory
        :param max_length:

        The size in bytes to cap the string at, or None [default] for no
        limit. This includes the NULL terminator.

        :param force: whether to ignore memory permissions
        """

        if max_length is not None:
            string = string[: max_length - 1]

        self.write_bytes(where, string + "\x00", force)

    def read_string(self, where: int, max_length: Optional[int] = None, force: bool = False) -> str:
        """
        Read a NUL-terminated concrete buffer from memory. Stops reading at first symbolic byte.

        :param where: Address to read string from
        :param max_length:
            The size in bytes to cap the string at, or None [default] for no
            limit.
        :param force: whether to ignore memory permissions
        :return: string read
        """
        s = io.BytesIO()
        while True:
            c = self.read_int(where, 8, force)

            if issymbolic(c) or c == 0:
                break

            if max_length is not None:
                if max_length == 0:
                    break
                max_length = max_length - 1
            s.write(Operators.CHR(c))
            where += 1
        return s.getvalue().decode()

    def push_bytes(self, data, force: bool = False):
        """
        Write `data` to the stack and decrement the stack pointer accordingly.

        :param data: Data to write
        :param force: whether to ignore memory permissions
        """
        self.STACK -= len(data)
        self.write_bytes(self.STACK, data, force)
        return self.STACK

    def pop_bytes(self, nbytes: int, force: bool = False):
        """
        Read `nbytes` from the stack, increment the stack pointer, and return
        data.

        :param nbytes: How many bytes to read
        :param force: whether to ignore memory permissions
        :return: Data read from the stack
        """
        data = self.read_bytes(self.STACK, nbytes, force=force)
        self.STACK += nbytes
        return data

    def push_int(self, value: int, force: bool = False):
        """
        Decrement the stack pointer and write `value` to the stack.

        :param value: The value to write
        :param force: whether to ignore memory permissions
        :return: New stack pointer
        """
        self.STACK -= self.address_bit_size // 8
        self.write_int(self.STACK, value, force=force)
        return self.STACK

    def pop_int(self, force: bool = False):
        """
        Read a value from the stack and increment the stack pointer.

        :param force: whether to ignore memory permissions
        :return: Value read
        """
        value = self.read_int(self.STACK, force=force)
        self.STACK += self.address_bit_size // 8
        return value

    #######################################
    # Decoder
    def _wrap_operands(self, operands):
        """
        Private method to decorate an Operand to our needs based on the
        underlying architecture.
        See :class:`~manticore.core.cpu.abstractcpu.Operand` class
        """
        raise NotImplementedError

    def decode_instruction(self, pc: int) -> Instruction:
        """
        This will decode an instruction from memory pointed by `pc`

        :param pc: address of the instruction
        """
        # No dynamic code!!! #TODO!
        # Check if instruction was already decoded
        if pc in self._instruction_cache:
            return self._instruction_cache[pc]

        text = b""

        # Read Instruction from memory
        exec_size = self.memory.max_exec_size(pc, self.max_instr_width)
        instr_memory = self.memory[pc : pc + exec_size]
        for i in range(exec_size):
            # This reads a byte from memory ignoring permissions
            # and concretize it if symbolic

            c = instr_memory[i]
            if issymbolic(c):
                # In case of fully symbolic memory, eagerly get a valid ptr
                if isinstance(self.memory, LazySMemory):
                    try:
                        vals = visitors.simplify_array_select(c)
                        c = bytes([vals[0]])
                    except visitors.ArraySelectSimplifier.ExpressionNotSimple:
                        self._publish("will_solve", self.memory.constraints, c, "get_value")
                        solved = SelectedSolver.instance().get_value(self.memory.constraints, c)
                        self._publish("did_solve", self.memory.constraints, c, "get_value", solved)
                        c = struct.pack("B", solved)
                elif isinstance(c, Constant):
                    c = bytes([c.value])
                else:
                    logger.error("Concretize executable memory %r %r", c, text)
                    raise ConcretizeMemory(
                        self.memory, address=pc, size=8 * self.max_instr_width, policy="INSTRUCTION"
                    )
            text += c

        # Pad potentially incomplete instruction with zeroes
        code = text.ljust(self.max_instr_width, b"\x00")

        try:
            # decode the instruction from code
            insn = self.disasm.disassemble_instruction(code, pc)
        except StopIteration as e:
            raise DecodeException(pc, code)

        # Check that the decoded instruction is contained in executable memory
        if insn.size > exec_size:
            logger.info("Trying to execute instructions from non-executable memory")
            raise InvalidMemoryAccess(pc, "x")

        insn.operands = self._wrap_operands(insn.operands)
        self._instruction_cache[pc] = insn
        return insn

    @property
    def instruction(self):
        if self._last_pc is None:
            return self.decode_instruction(self.PC)
        else:
            return self.decode_instruction(self._last_pc)

    #######################################
    # Execute
    def canonicalize_instruction_name(self, instruction):
        """
        Get the semantic name of an instruction.
        """
        raise NotImplementedError

    def execute(self):
        """
        Decode, and execute one instruction pointed by register PC
        """
        curpc = self.PC
        if self._delayed_event:
            self._publish_instruction_as_executed(self.decode_instruction(self._last_pc))
            self._delayed_event = False

        if issymbolic(curpc):
            raise ConcretizeRegister(self, "PC", policy="ALL")
        if not self.memory.access_ok(curpc, "x"):
            raise InvalidMemoryAccess(curpc, "x")

        self._publish("will_decode_instruction", curpc)

        insn = self.decode_instruction(curpc)
        self._last_pc = self.PC

        self._publish("will_execute_instruction", self._last_pc, insn)

        # FIXME (theo) why just return here?
        # hook changed PC, so we trust that there is nothing more to do
        if insn.address != self.PC:
            self._last_executed_pc = insn.address
            return

        name = self.canonicalize_instruction_name(insn)

        if logger.level == logging.DEBUG:
            logger.debug(self.render_instruction(insn))
            for l in self.render_registers():
                register_logger.debug(l)

        try:
            if self._concrete and "SYSCALL" in name:
                self.emu.sync_unicorn_to_manticore()
            if self._concrete and "SYSCALL" not in name:
                self.emulate(insn)
                if self.PC == self._break_unicorn_at:
                    logger.debug("Switching from Unicorn to Manticore")
                    self._break_unicorn_at = None
                    self._concrete = False
            else:
                implementation = getattr(self, name, None)

                if implementation is not None:
                    implementation(*insn.operands)

                else:
                    text_bytes = " ".join("%02x" % x for x in insn.bytes)
                    logger.warning(
                        "Unimplemented instruction: 0x%016x:\t%s\t%s\t%s",
                        insn.address,
                        text_bytes,
                        insn.mnemonic,
                        insn.op_str,
                    )
                    self.backup_emulate(insn)
        except (Interruption, Syscall) as e:
            self._delayed_event = True
            raise e
        else:
            self._publish_instruction_as_executed(insn)

    # FIXME(yan): In the case the instruction implementation invokes a system call, we would not be able to
    # publish the did_execute_instruction event from here, so we capture and attach it to the syscall
    # exception for the platform to emit it for us once the syscall has successfully been executed.
    def _publish_instruction_as_executed(self, insn):
        """
        Notify listeners that an instruction has been executed.
        """
        self._last_executed_pc = self._last_pc
        self._icount += 1
        self._publish("did_execute_instruction", self._last_pc, self.PC, insn)

    def emulate(self, insn):
        """
        Pick the right emulate function (maintains API compatiblity)

        :param insn: single instruction to emulate/start emulation from
        """

        if self._concrete:
            self.concrete_emulate(insn)
        else:
            self.backup_emulate(insn)

    def concrete_emulate(self, insn):
        """
        Start executing in Unicorn from this point until we hit a syscall or reach break_unicorn_at

        :param capstone.CsInsn insn: The instruction object to emulate
        """

        if not self.emu:
            self.emu = ConcreteUnicornEmulator(self)
        if self.emu._stop_at is None:
            self.emu.write_backs_disabled = False
            self.emu._stop_at = self._break_unicorn_at
            self.emu.load_state_from_manticore()
        try:
            self.emu.emulate(insn)
        except unicorn.UcError as e:
            if e.errno == unicorn.UC_ERR_INSN_INVALID:
                text_bytes = " ".join("%02x" % x for x in insn.bytes)
                logger.error(
                    "Unimplemented instruction: 0x%016x:\t%s\t%s\t%s",
                    insn.address,
                    text_bytes,
                    insn.mnemonic,
                    insn.op_str,
                )
            raise InstructionEmulationError(str(e))

    def backup_emulate(self, insn):
        """
        If we could not handle emulating an instruction, use Unicorn to emulate
        it.

        :param capstone.CsInsn instruction: The instruction object to emulate
        """

        if not hasattr(self, "backup_emu"):
            self.backup_emu = UnicornEmulator(self)
        try:
            self.backup_emu.emulate(insn)
        except unicorn.UcError as e:
            if e.errno == unicorn.UC_ERR_INSN_INVALID:
                text_bytes = " ".join("%02x" % x for x in insn.bytes)
                logger.error(
                    "Unimplemented instruction: 0x%016x:\t%s\t%s\t%s",
                    insn.address,
                    text_bytes,
                    insn.mnemonic,
                    insn.op_str,
                )
            raise InstructionEmulationError(str(e))
        finally:
            # We have been seeing occasional Unicorn issues with it not clearing
            # the backing unicorn instance. Saw fewer issues with the following
            # line present.
            del self.backup_emu

    def render_instruction(self, insn=None):
        try:
            insn = self.instruction
            return f"INSTRUCTION: 0x{insn.address:016x}:\t{insn.mnemonic}\t{insn.op_str}"
        except Exception as e:
            return "{can't decode instruction}"

    def render_register(self, reg_name):
        result = ""

        value = self.read_register(reg_name)

        if issymbolic(value):
            value = str(value)  # coerce the value into a string
            aux = f"{reg_name:3s}: {value:16s}"
            result += aux
        elif isinstance(value, int):
            result += f"{reg_name:3s}: 0x{value:016x}"
        else:
            result += f"{reg_name:3s}: {value!r}"
        return result

    def render_registers(self):
        # FIXME add a context manager at utils that look for all Signal
        # backup, null, use, then restore the list.
        # will disabled_signals(self):
        #    return map(self.render_register, self._regfile.canonical_registers)
        return map(self.render_register, sorted(self._regfile.canonical_registers))

    # Generic string representation
    def __str__(self):
        """
        Returns a string representation of cpu state

        :rtype: str
        :return: name and current value for all the registers.
        """
        result = f"{self.render_instruction()}\n"
        result += "\n".join(self.render_registers())
        return result


# Instruction decorators


def instruction(old_method):
    # This should decorate every instruction implementation
    @wraps(old_method)
    def new_method(cpu, *args, **kw_args):
        cpu.PC += cpu.instruction.size
        return old_method(cpu, *args, **kw_args)

    new_method.old_method = old_method
    return new_method
